#!/usr/bin/env python3
"""
Custom exceptions.
"""


import sys


THIS_MODULE = sys.modules[__name__]


class LoaderError(Exception):
    """
    Custom exception base class.
    """


class ExpectedExceptionContext:
    """
    Context manager to convert expected exceptions other exception types.
    """

    # Predefined map of commonly expected exceptions to custom exception
    # subclasses.
    EXCEPTION_MAP = {}

    @classmethod
    def map_exception(cls, ex_from, ex_to):
        """
        Add an exception pair to the exception map.

        Args:
            ex_from:
                The exception to map.

            ex_to:
                The exception to which ex_from should be mapped.
        """
        for ex in (ex_from, ex_to):
            if not issubclass(ex, Exception):
                raise ValueError(f"{ex} is not an instance of Exception")
        cls.EXCEPTION_MAP[ex_from] = ex_to

    def __init__(self, *expected, error_msg=None):
        """
        Args:
            *expected:
                The expected exceptions to transform into custom exceptions.
                These must be exception classes that are already in
                EXCEPTION_MAP.

            error_msg:
                An error message to prepend to the caught exception when
                instantiating the custom exception. It should provide context
                for the exception when it is displayed to the user.
        """
        expected = tuple(expected)
        for exception in expected:
            if exception not in self.EXCEPTION_MAP:
                raise ValueError(
                    f"{exception} is not currently mapped to any other exception"
                )
        self._expected = expected
        self.error_msg = error_msg

    def __enter__(self):
        return self._expected

    def __exit__(self, typ, value, traceback):
        if typ is not None:
            for exp_type in self._expected:
                if issubclass(typ, exp_type):
                    if self.error_msg:
                        arg = f"{self.error_msg}: {value}"
                    else:
                        arg = value
                    raise self.EXCEPTION_MAP[exp_type](arg) from value


def _declare_custom_exceptions():
    """
    Create custom subclasses for mapping common exceptions and add them to
    ExpectedExceptionContext exception map.
    """
    for std_exc in (KeyError, OSError, ValueError):
        name = std_exc.__name__
        custom_exc = type(f"Loader{name}", (std_exc,), {})
        custom_exc.__doc__ = f"Custom {name} exception."
        setattr(THIS_MODULE, custom_exc.__name__, custom_exc)
        ExpectedExceptionContext.EXCEPTION_MAP[std_exc] = custom_exc


_declare_custom_exceptions()



#!/usr/bin/env python3
"""
Regular expression types and functions.
"""

import enum
import fnmatch
import logging
import re


LOGGER = logging.getLogger(__name__)


@enum.unique
class PatternType(enum.Enum):
    """
    Recognized pattern types.
    """

    REGEX = enum.auto()
    GLOB = enum.auto()
    LITERAL = enum.auto()

    @classmethod
    def from_str(cls, arg):
        """
        Convert a string to a PatternType.
        """
        arg = str(arg).upper()
        for ptype in cls:
            if ptype.name == arg:
                return ptype
        raise ValueError(f"Unrecognized pattern type: {arg}")


class RegexPattern:  # pylint: disable=too-few-public-methods
    """
    Regular expression pattern.
    """

    TYPE = PatternType.REGEX

    def __init__(self, pattern, flags=0):
        if not isinstance(pattern, re.Pattern):
            pattern = re.compile(pattern, flags=flags)
        self.regex = pattern
        self.pattern = self.regex.pattern

    def fullmatch(self, *args, **kwargs):
        """
        Wrapper around re.Pattern.fullmatch
        """
        return self.regex.fullmatch(*args, **kwargs)


class GlobPattern(RegexPattern):  # pylint: disable=too-few-public-methods
    """
    Glob pattern.
    """

    TYPE = PatternType.GLOB

    def __init__(self, pattern):
        super().__init__(fnmatch.translate(pattern))
        self.pattern = pattern


class LiteralPattern(RegexPattern):  # pylint: disable=too-few-public-methods
    """
    Literal pattern.
    """

    TYPE = PatternType.LITERAL

    def __init__(self, pattern):
        super().__init__(re.escape(pattern))
        self.pattern = pattern


def get_regex(pattern, pattern_type=PatternType.REGEX):
    """
    Get the regular expression corresponding to the given pattern.

    Args:
        pattern:
            A string containing a pattern of the specified type that should be
            converted to a regular expression.

        pattern_type:
            An instance of :py:class:`~.PatternType` or an equivalent string.

    Returns:
        An re.Pattern regular expression object.
    """
    if isinstance(pattern, RegexPattern):
        return pattern

    if isinstance(pattern_type, str):
        pattern_type = PatternType.from_str(pattern_type)

    if pattern_type is PatternType.LITERAL:
        return LiteralPattern(pattern)

    if pattern_type is PatternType.GLOB:
        return GlobPattern(pattern)

    if pattern_type is PatternType.REGEX:
        return RegexPattern(pattern)

    raise ValueError(f"Unrecogznied pattern type: {pattern_type}")


def get_extension_pattern_for_filepath(ext, escape=True):
    r"""
    Get a regular expression pattern for filepaths that end with the given
    extension. This will exclude URIs, including file URIs.

    Args:
        ext:
            The extention to recognize, e.g. ".txt".

        escape:
            If True, escape the extension for the pattern. This can be set to
            false whena pre-escaped pattern is passed in, e.g. r"\.[tc]sv".

    Returns:
        A RegexPattern or a subclass thereof.
    """
    if escape:
        ext = re.escape(ext)
    return RegexPattern(rf"^(?!\w+://).*{ext}$")



#!/usr/bin/env python3
"""
Test text loaders.
"""

import unittest

from dalood.exception import LoaderValueError
from dalood.loader.text import TextFileLoader, TextUrlLoader
from dalood.manager import Manager


from test_dalood.utils import tmp_file


class TestText(unittest.TestCase):
    """
    Test text loading.
    """

    def test_text_file_load(self):
        """
        Text is loaded from files.
        """
        man = Manager()
        man.register_loader(r"^.*\.txt$", TextFileLoader())
        with tmp_file() as (text, _, path):
            self.assertEqual(text, man.get(path))

    def test_text_file_mtime(self):
        """
        Text file modification time is correct.
        """
        man = Manager()
        man.register_loader(r"^.*\.txt$", TextFileLoader())
        with tmp_file() as (_, mtime, path):
            self.assertEqual(mtime, man.get_mtime(path))

    def test_uri_file_load(self):
        """
        Text is loaded from URIs.
        """
        man = Manager()
        man.register_loader(r"^file://.*$", TextUrlLoader())
        with tmp_file() as (text, _, path):
            self.assertEqual(text, man.get(path.as_uri()))

    def test_text_uri_mtime(self):
        """
        Text URI modification time is correct.
        """
        man = Manager()
        man.register_loader(r"^file://.*$", TextUrlLoader())
        with tmp_file() as (_, mtime, path):
            self.assertEqual(mtime.replace(microsecond=0), man.get_mtime(path.as_uri()))

    def test_patterns(self):
        """
        Text file loader registers its own patterns.
        """
        man = Manager()
        loader = TextFileLoader()
        with tmp_file() as (text, _, path):
            with self.assertRaises(LoaderValueError):
                man.get(path.as_uri())

            loader.register_patterns(man)
            self.assertEqual(text, man.get(path))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
YAML data loaders.
"""

import yaml

from ..exception import ExpectedExceptionContext
from ..regex import get_extension_pattern_for_filepath
from .file import FileLoaderBase
from .url import UrlLoaderBase


ExpectedExceptionContext.map_exception(yaml.YAMLError, ValueError)


class YAMLFileLoader(FileLoaderBase):
    """
    YAML file data loader.
    """

    def __init__(self, encoding="utf-8"):
        """
        Args:
            encoding:
                The file encoding.
        """
        super().__init__("r", encoding=encoding)

    def load(self, src):
        with self.stream(src, yaml.YAMLError) as handle:
            return yaml.safe_load(handle)

    @property
    def patterns(self):
        yield get_extension_pattern_for_filepath(r"\.ya?ml", escape=False)


class YAMLUrlLoader(UrlLoaderBase):
    """
    YAML URL data loader.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load(self, src):
        with self.stream(src, yaml.YAMLError) as resp:
            return yaml.safe_load(resp.raw)



#!/usr/bin/env python3
"""
Test Pandas DataFrame loaders.
"""

import pathlib
import sqlite3
import tempfile
import unittest

import pandas as pd

from dalood.exception import LoaderValueError
from dalood.loader.pandas import DataFrameCSVLoader, DataFrameSQLLoader
from dalood.manager import Manager


from test_dalood.utils import tmp_file


CSV_EXT = ".csv"


class TestDataFrameCSV(unittest.TestCase):
    """
    Test Pandas DataFrame loading from CSV files.
    """

    def setUp(self):
        self.data = pd.DataFrame({"number": [1, 2, 3], "letter": ["a", "b", "c"]})
        self.tmp_file_ctx = tmp_file(
            text=self.data.to_csv(index=False), extension=CSV_EXT
        )
        self.man = Manager()
        self.man.register_loader(
            f"*{CSV_EXT}", DataFrameCSVLoader(), pattern_type="glob"
        )

    def test_csv_file_load(self):
        """
        Pandas DataFrame is loaded from file.
        """
        with self.tmp_file_ctx as (_, _, path):
            self.assertTrue(self.data.equals(self.man.get(path)))

    def test_same_object(self):
        """
        Manager returns the same object on subsequent calls.
        """
        with self.tmp_file_ctx as (_, _, path):
            self.assertIs(self.man.get(path), self.man.get(path))

    def test_patterns(self):
        """
        Pandas file loader registers its own patterns.
        """
        man = Manager()
        loader = DataFrameCSVLoader()
        with self.tmp_file_ctx as (_, _, path):
            with self.assertRaises(LoaderValueError):
                man.get(path.as_uri())

            loader.register_patterns(man)
            self.assertTrue(self.data.equals(man.get(path)))


class TestDataFrameSQL(unittest.TestCase):
    """
    Test Pandas DataFrame loading from SQL databases.
    """

    def setUp(self):
        self.tmp_dir_obj = (
            tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        )
        self.db_path = pathlib.Path(self.tmp_dir_obj.name) / "db.sqlite3"
        self.data = pd.DataFrame({"number": [1, 2, 3], "letter": ["a", "b", "c"]})
        self.table_name = "test"

        conn = sqlite3.connect(self.db_path)
        self.data.to_sql(self.table_name, conn, index=False)

        self.man = Manager()
        self.loader = DataFrameSQLLoader(con=conn)
        self.loader.map("sqltest", f"SELECT * FROM {self.table_name}")
        self.loader.register_patterns(self.man)

    def tearDown(self):
        self.tmp_dir_obj.cleanup()

    def test_sql_load(self):
        """
        Pandas DataFrame is loaded from SQL.
        """
        self.assertTrue(self.data.equals(self.man.get("sqltest")))

    def test_map(self):
        """
        New sources are mapped to new SQL statements.
        """
        self.loader.map("sqltest", f"SELECT letter FROM {self.table_name}")
        self.assertTrue(
            self.data.drop(["number"], axis=1).equals(self.man.get("sqltest"))
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Data wrapper class.
"""

import datetime
import logging

LOGGER = logging.getLogger(__name__)


class DataWrapper:
    """
    Wrapper around loaded data that stores that data and a loader reference, and
    tracks loading and access times.
    """

    def __init__(self, loader, src):
        self.loader = loader
        self.src = src
        self.data = None
        self.load_time = None
        self.access_time = None

    @staticmethod
    def _now():
        """
        Get the current datetime in the UTC timezone.
        """
        return datetime.datetime.now(datetime.timezone.utc)

    def load(self):
        """
        Load the data for the given source.

        Args:
            src:
                The data source (e.g. a file or URI, or whatever else the
                matching loader can handle).
        """
        self.data = self.loader.load(self.src)
        self.load_time = self._now()
        self.access_time = self.load_time

    def refresh(self):
        """
        Reload data if the loader reports that the data has changed since it was
        last loaded.
        """
        mtime = self.loader.get_mtime(self.src)
        if mtime is not None and mtime > self.load_time:
            LOGGER.debug("Refreshing %s", self.src)
            self.load()

    def access(self):
        """
        Access the data. It will be loaded first if necessary.
        """
        if self.load_time is None:
            self.load()
        self.access_time = self._now()
        return self.data



#!/usr/bin/env python3
"""
Pandas dataframe loaders
"""

import logging

import pandas as pd

from ..exception import ExpectedExceptionContext
from ..regex import get_extension_pattern_for_filepath
from .file import FileLoaderBase
from .map import MapLoaderBase


LOGGER = logging.getLogger(__name__)


class DataFrameCSVLoader(FileLoaderBase):
    """
    CSV & TSV file dataframe loader.
    """

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs:
                Keyword arguments for pandas.read_csv().
        """
        super().__init__()
        self.kwargs = kwargs

    def load(self, src):
        src = str(src)
        LOGGER.debug("Loading Pandas dataframe from %s", src)
        with ExpectedExceptionContext(OSError, ValueError):
            return pd.read_csv(src, **self.kwargs)

    @property
    def patterns(self):
        yield get_extension_pattern_for_filepath(r"\.[ct]sv", escape=False)


class DataFrameSQLLoader(MapLoaderBase):
    """
    Text URL data loader.
    """

    def __init__(self, mapping=None, **kwargs):
        """
        Args:
            mapping:
                A dict mapping the source arguments to SQL statements. When a
                source is requested, the mapped SQL statement will be passed
                through to pandas.read_sql() along with the keyword arguments in
                kwargs.

            **kwargs:
                Keyword arguments for pandas.read_sql().
        """
        super().__init__(mapping=mapping)
        self.kwargs = kwargs

    def load(self, src):
        src = str(src)
        LOGGER.debug("Loading Pandas dataframe from %s", src)
        with ExpectedExceptionContext(OSError, ValueError, KeyError):
            sql = self.mapping[src]
            return pd.read_sql(sql, **self.kwargs)



#!/usr/bin/env python3
"""
Test YAML loaders.
"""

import unittest

import yaml

from dalood.exception import LoaderValueError
from dalood.loader.yaml import YAMLFileLoader, YAMLUrlLoader
from dalood.manager import Manager


from test_dalood.utils import tmp_file


YAML_EXT = ".yaml"


def _get_yaml():
    """
    Get data and its YAML representation.
    """
    data = {"foo": 1, "bar": 2, "items": [3, 4, 5], "bool": True}
    return data, yaml.dump(data)


class TestYAML(unittest.TestCase):
    """
    Test YAML loading.
    """

    def setUp(self):
        self.data, self.yaml_txt = _get_yaml()
        self.tmp_file_ctx = tmp_file(text=self.yaml_txt, extension=YAML_EXT)

    def test_yaml_file_load(self):
        """
        YAML is loaded from files.
        """
        man = Manager()
        man.register_loader(r"^.*\.yaml$", YAMLFileLoader())
        with self.tmp_file_ctx as (_, _, path):
            self.assertEqual(self.data, man.get(path))

    def test_uri_file_load(self):
        """
        YAML is loaded from URIs.
        """
        man = Manager()
        man.register_loader(r"^file://.*$", YAMLUrlLoader())
        with self.tmp_file_ctx as (_, _, path):
            self.assertEqual(self.data, man.get(path.as_uri()))

    def test_patterns(self):
        """
        YAML file loader registers its own patterns.
        """
        man = Manager()
        loader = YAMLFileLoader()
        with self.tmp_file_ctx as (_, _, path):
            with self.assertRaises(LoaderValueError):
                man.get(path.as_uri())

            loader.register_patterns(man, prioritize=True)
            self.assertEqual(self.data, man.get(path))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
JSON data loaders.
"""

import json

from ..exception import ExpectedExceptionContext
from ..regex import get_extension_pattern_for_filepath
from .file import FileLoaderBase
from .url import UrlLoaderBase


ExpectedExceptionContext.map_exception(json.JSONDecodeError, ValueError)


class JSONFileLoader(FileLoaderBase):
    """
    JSON file data loader.
    """

    def __init__(self):
        super().__init__("rb")

    def load(self, src):
        with self.stream(src, json.JSONDecodeError) as handle:
            return json.load(handle)

    @property
    def patterns(self):
        yield get_extension_pattern_for_filepath(".json")


class JSONUrlLoader(UrlLoaderBase):
    """
    JSON URL data loader.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load(self, src):
        with self.stream(src) as resp:
            return resp.json()



#!/usr/bin/env python3
"""
File manager class.
"""

import datetime
import functools
import inspect
import logging
from collections import OrderedDict

from .data_wrapper import DataWrapper
from .exception import LoaderValueError  # pylint: disable=no-name-in-module
from .regex import PatternType, get_regex


LOGGER = logging.getLogger(__name__)


class Manager:
    """
    File manager for loading files on demand and managing memory.

    Attributes:
        loaders:
            An OrderedDict mapping argument regular expressions to loaders.
    """

    def __init__(self):
        self.loaders = OrderedDict()
        self._cache = {}

    def register_loader(
        self, pattern, loader, *, prioritize=False, pattern_type=PatternType.REGEX
    ):
        """
        Register a loader to handle a pattern.

        Args:
            pattern:
                Either a string, in which case the pattern type is determined by
                the pattern_type argument, or an instance of RegexPattern or one
                of its subclasses.

            loader:
                A subclass of the :py:class:`~.loader.base.LoaderBase`.

            prioritize:
                If True, insert this regex at the start of the list of registred
                regexes, otherwise it will be inserted at the end.

            pattern_type:
                The pattern type, either as an instance of
                :py:class:`~.regex.PatternType` or a string that is
                recognized by :py:meth:`.regex.PatternType.from_str`
                (e.g. "glob", "literal").
        """
        regex = get_regex(pattern, pattern_type=pattern_type)
        LOGGER.debug("Registering loader %s with regex %s", loader, regex)
        self.loaders[regex] = loader
        if prioritize:
            self.loaders.move_to_end(regex, last=False)

    def get_loader(self, src):
        """
        Get the loader that matches the source.

        Args:
            src:
                The data source (e.g. a file or URI, or whatever else the
                matching loader can handle).

        Raises:
            ValueError:
                The given source is not matched by any registered regular
                expressions.
        """
        src = str(src)
        try:
            return self._cache[src].loader
        except KeyError:
            pass
        for regex, loader in self.loaders.items():
            if regex.fullmatch(src):
                return loader
        raise LoaderValueError(
            f'None of the registered regular expressions match the source "{src}": '
            "unable to find a corresponding data loader"
        )

    def clear_cache(
        self,
        pattern=None,
        age=None,
        by_access_time=False,
        pattern_type=PatternType.REGEX,
    ):
        """
        Clear cached data from memory.

        Args:
            pattern:
                Same as for :py:meth:`~.Manager.register_loader`. If
                given, only cached items matching this pattern will be cleared.

            pattern_type:
                Same as for :py:meth:`~.Manager.register_loader`.

            age:
                An optional datetime.timedelta object (or a dict of keyword
                arguments that can be passed to datetime.timedelta(). If the
                data has been loaded for longer than this time, it will be
                cleared.

            by_access_time:
                If True, clear data by last access time when clearing by age,
                otherwise clear by load time.
        """
        if isinstance(age, dict):
            age = datetime.timedelta(**age)
        elif not (isinstance(age, datetime.timedelta) or age is None):
            raise ValueError(f"Invalid argument type for age: {type(age)}")

        time_attr = "access_time" if by_access_time else "load_time"

        if pattern is None:
            if age is None:
                self._cache.clear()
            else:
                now = datetime.datetime.now(datetime.timezone.utc)
                for src, item in list(self._cache.items()):
                    if (now - getattr(item, time_attr)) > age:
                        del self._cache[src]
            return

        regex = get_regex(pattern, pattern_type=pattern_type)
        for src, item in list(self._cache.items()):
            if regex.fullmatch(src):
                if age is None or (now - getattr(item, time_attr)) > age:
                    del self._cache[src]

    def refresh(self, pattern=None, pattern_type=PatternType.REGEX):
        """
        Refresh data for which the loader reports a source modification since
        the time the data was loaded.

        Args:
            pattern:
                Same as for :py:meth:`~.Manager.register_loader`.  If
                given, only cached items matching this pattern will be
                refreshed.

            pattern_type:
                Same as for :py:meth:`~.Manager.register_loader`.
        """
        wrappers = self._cache.values()
        if pattern is not None:
            regex = get_regex(pattern, pattern_type=pattern_type)
            wrappers = (
                wrapper for wrapper in wrappers if regex.fullmatch(str(wrapper.src))
            )
        for wrapper in wrappers:
            wrapper.refresh()

    def _get_wrapper(self, src):
        """
        Get the wrapper for the given source.

        Args:
            src:
                The data source.

        Returns:
            The wrapper for this source.
        """
        str_src = str(src)
        try:
            wrapper = self._cache[str_src]
        except KeyError:
            LOGGER.debug("%s is not in cache.", str_src)
            loader = self.get_loader(src)
            wrapper = DataWrapper(loader, src)
            self._cache[str(src)] = wrapper
            return wrapper
        LOGGER.debug("Found %s in cache.", str_src)
        return wrapper

    def get(self, src, refresh=False, reload=False):
        """
        Get the data from the given source. The data will be loaded via the
        loader if necessary.

        Args:
            src:
                The data source.

            refresh:
                If True, reload data if the reported modification time is newer
            H    than the loaded data.

            reload:
                If True, force a reload.

        Returns:
            The data.
        """
        wrapper = self._get_wrapper(src)
        if reload:
            LOGGER.debug("Forcing reload of %s.", src)
            wrapper.load()
        elif refresh:
            wrapper.refresh()
        return wrapper.access()

    def get_mtime(self, src):
        """
        Get the modification time of a source.

        Args:
            src:
                The data source.

        Returns:
            The modification time as a datetime.datetime object, or None.
        """
        str_src = str(src)
        try:
            loader = self._cache[str_src].loader
        except KeyError:
            loader = self.get_loader(src)
        return loader.get_mtime(src)

    def __iter__(self):
        """
        Iterator over cached sources.
        """
        yield from self._cache.keys()


def _get_fixed_pattern_type_method(name, pattern_type):
    def new_method(self, pattern, *args, **kwargs):
        method = getattr(self, name)
        kwargs["pattern_type"] = pattern_type
        return method(pattern, *args, **kwargs)

    return new_method


def _add_convenience_methods():
    """
    Programmatically add convenience methods for Manager methods that accept
    different pattern types.
    """
    for name, method in inspect.getmembers(Manager, predicate=inspect.isfunction):
        signature = inspect.signature(method)
        params = signature.parameters
        if "pattern" in params and "pattern_type" in params:
            new_params = OrderedDict(
                item for item in params.items() if item[0] != "pattern_type"
            )
            new_params["pattern"] = new_params["pattern"].replace(
                name="pattern",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=inspect.Parameter.empty,
            )
            new_sig = signature.replace(parameters=new_params.values())
            for pattern_type in PatternType:
                new_method = _get_fixed_pattern_type_method(name, pattern_type)
                new_method.__doc__ = (
                    f"Wrapper around :py:meth:`~{__name__}.Manager.{name}` "
                    f"for {pattern_type.name} patterns."
                )
                new_method.__name__ = f"{name}_with_{pattern_type.name.lower()}_pattern"
                new_method.__signature__ = new_sig
                setattr(Manager, new_method.__name__, new_method)


_add_convenience_methods()



#!/usr/bin/env python3
"""
Base class for loaders that map sources to data via a user-defined map.
"""

import logging


from ..regex import LiteralPattern
from .base import LoaderBase


LOGGER = logging.getLogger(__name__)


class MapLoaderBase(LoaderBase):
    """
    Base class for loaders that map sources to data via a user-defined map.
    """

    def __init__(self, mapping=None):
        """
        Args:
            mapping:
                A dict mapping the source arguments to user-defined data. How
                that data is used will be determined by the implementation of
                load and get_mtime in derived classes. For example, for a loader
                that loads Pandas DataFrames from a database, the mapping could
                map source arguments to SQL statements that are passed through
                to pandas.read_sql.
        """
        super().__init__()
        if mapping is None:
            mapping = {}
        self.mapping = mapping

    def map(self, src, value):
        """
        Map a source to a value in the internal mapping.

        Args:
            src:
                The source to map.

            value:
                The value to which to map the source.
        """
        self.mapping[src] = value

    @property
    def patterns(self):
        for key in self.mapping:
            yield LiteralPattern(key)



#!/usr/bin/env python3
"""
Data loader base class.
"""

from abc import ABC, abstractmethod


class LoaderBase(ABC):
    """
    Base class for loading data from URIs or filepaths.
    """

    @abstractmethod
    def load(self, src):
        """
        Load data from the given source.

        Args:
            src:
                The data source. For some loaders this may simply be a file path
                or URI. For others it may be an arbitrary string that only has
                meaning to the loader. For example, some loaders can map
                user-defined strings to pre-defined static data, or SQL
                statements that can be used to retrieve a Pandas DataFrame from
                an open database connection.

        Returns:
            The loaded data in a form appropriate for the loader.
        """

    def get_mtime(self, src):  # pylint: disable=unused-argument
        """
        Attempt to determine the last modification time of the source.

        Args:
            src:
                The data source (e.g. a file path or URI).

        Returns:
            A datetime.datetime object representing the last known modification
            time, or None if the time cannot be determined.
        """
        return None

    @property
    def patterns(self):
        r"""
        An iterator over 2-tuples of patterns and their PatternType associated
        with this loader. For example, a text file loader might return the
        2-tuple (r"^(?!https?://).*\.txt$", PatternType.REGEX) to load local
        files by the ".txt" extenstion.
        """
        yield from []

    def register_patterns(self, manager, prioritize=False):
        """
        Register this loader with a manager for each of its common patterns.

        Args:
            manager:
                An instance of Manager.

            prioritize:
                If True, ensure that this loaders patterns take precedence over
                existing patterns in the manager.
        """
        if prioritize:
            for pattern in reversed(list(self.patterns)):
                manager.register_loader(pattern, self, prioritize=prioritize)
        else:
            for pattern in self.patterns:
                manager.register_loader(pattern, self)



#!/usr/bin/env python3
"""
Base class for other URL loaders.
"""

import contextlib
import datetime
import logging

from requests.exceptions import RequestException
from requests_file_adapter import get_session

from ..exception import ExpectedExceptionContext, LoaderOSError
from .base import LoaderBase


LOGGER = logging.getLogger(__name__)


ExpectedExceptionContext.map_exception(RequestException, LoaderOSError)


class UrlLoaderBase(LoaderBase):
    """
    Base class for loading data from URLs. This will check modification times
    via a HEAD request.
    """

    def __init__(self, timeout=5):
        """
        Args:
            timeout:
                Timeout parameter passed through to requests.
        """
        self.timeout = timeout

    @contextlib.contextmanager
    def stream(self, src, *exceptions):
        """
        Context manager for accessing the data raw stream. It returns the
        response through which the content can be accessed.
        """
        LOGGER.debug("Getting %s", src)
        session = get_session()
        with ExpectedExceptionContext(RequestException, *exceptions):
            resp = session.get(src, timeout=self.timeout, stream=True)
            yield resp
        resp.close()

    def get_mtime(self, src):
        time_fmt = "%a, %d %b %Y %H:%M:%S %Z"
        session = get_session()
        with ExpectedExceptionContext(RequestException):
            resp = session.head(src, timeout=self.timeout)
        try:
            last_mod = resp.headers["last-modified"]
        except KeyError:
            return None
        mtime = datetime.datetime.strptime(last_mod, time_fmt)
        # strptime seems to return naive datetime objects without tzinfo
        if mtime.tzinfo is None and last_mod.endswith(" GMT"):
            mtime = mtime.replace(tzinfo=datetime.timezone.utc)
        return mtime



#!/usr/bin/env python3
"""Test utilities."""


import contextlib
import datetime
import pathlib
import tempfile


@contextlib.contextmanager
def tmp_file(text=None, extension=".txt"):
    """
    Create a temporary file in a temporary directory.

    Returns:
        A 3-tuple with the file content, the file modification time and the
        path.
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir) / f"tmp{extension}"
        if text is None:
            text = "placeholder text"
        tmp_path.write_text(text)
        mtime = datetime.datetime.fromtimestamp(
            tmp_path.stat().st_mtime, datetime.timezone.utc
        )
        yield text, mtime, tmp_path



#!/usr/bin/env python3
'''
Version file automatically generated by setuptools-scm via hatch-vcs.
https://pypi.org/project/setuptools-scm/
https://github.com/ofek/hatch-vcs
Do not track this file and do not edit it manually.
'''
VERSION = '2024.7'
VERSION_TUPLE = (2024, 7)



#!/usr/bin/env python3
"""
Loader for objects in memory.
"""

import logging


from ..exception import ExpectedExceptionContext
from .map import MapLoaderBase


LOGGER = logging.getLogger(__name__)


class MemoryLoader(MapLoaderBase):
    """
    Loader for object in memory. This simply provided a uniform API for
    accessing objects already in memory such as those passed in by the user. In
    this case, the mapping of the parent MapLoaderBase class will map source
    arguments directly to returned data.
    """

    def load(self, src):
        with ExpectedExceptionContext(KeyError):
            return self.mapping[src]



#!/usr/bin/env python3
"""
Text data loaders.
"""

import logging

from ..regex import get_extension_pattern_for_filepath
from .file import FileLoaderBase
from .url import UrlLoaderBase


LOGGER = logging.getLogger(__name__)


class TextFileLoader(FileLoaderBase):
    """
    Text file data loader.
    """

    def __init__(self, encoding="utf-8"):
        """
        Args:
            encoding:
                The file encoding.
        """
        super().__init__("r", encoding=encoding)

    def load(self, src):
        with self.stream(src) as handle:
            return handle.read()

    @property
    def patterns(self):
        yield get_extension_pattern_for_filepath(".txt")


class TextUrlLoader(UrlLoaderBase):
    """
    Text URL data loader.
    """

    def __init__(self, *args, encoding="utf-8", **kwargs):
        """
        Args:
            encoding:
                The stream encoding.
        """
        super().__init__(*args, **kwargs)
        self.encoding = encoding

    def load(self, src):
        with self.stream(src) as resp:
            text = resp.content
            if self.encoding is not None:
                return text.decode(self.encoding)
            return text



#!/usr/bin/env python3
"""
Test map and memory loader.
"""

import json
import unittest

from dalood.loader.memory import MemoryLoader
from dalood.manager import Manager


class TestMemory(unittest.TestCase):
    """
    Test memory loading.
    """

    def setUp(self):
        self.man = Manager()
        self.data = {
            "test1": ("a", 1),
            "test2": ("b", 2),
        }

    def test_init_mapping(self):
        """
        Mapping is recognized when passed to __init__.
        """
        loader = MemoryLoader(mapping=self.data)
        loader.register_patterns(self.man)
        for name, data in self.data.items():
            with self.subTest(name=name):
                self.assertEqual(data, self.man.get(name))

    def test_map_method(self):
        """
        Mapping via map function works.
        """
        loader = MemoryLoader()
        for name, data in self.data.items():
            loader.map(name, data)
        loader.register_patterns(self.man)
        for name, data in self.data.items():
            with self.subTest(name=name):
                self.assertEqual(data, self.man.get(name))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Base class for other file loaders.
"""

import contextlib
import datetime
import logging
import pathlib

from ..exception import ExpectedExceptionContext
from .base import LoaderBase


LOGGER = logging.getLogger(__name__)


class FileLoaderBase(LoaderBase):
    """
    Base class for loading data from files.
    """

    DEFAULT_ENCODING = "utf-8"

    def __init__(self, *args, **kwargs):
        """
        Args:
            *args:
                Positional arguments passed through to pathlib.open after the
                path.

            **kwargs:
                Keyword arguments passed through to pathlib.open.
        """
        self.args = args
        self.kwargs = kwargs

    @contextlib.contextmanager
    def stream(self, src, *exceptions):
        """
        Context manager for accessing the open file buffer.
        """
        path = pathlib.Path(src).resolve()
        args = self.args if self.args else ["rb"]
        kwargs = {**self.kwargs}
        if args[0] == "r":
            kwargs.setdefault("encoding", self.DEFAULT_ENCODING)

        LOGGER.debug("Opening %s with %s, %s", path, args, kwargs)
        with ExpectedExceptionContext(
            OSError, *exceptions, error_msg=f"Failed to open {path}"
        ):
            with path.open(  # pylint: disable=unspecified-encoding
                *args, **kwargs
            ) as handle:
                yield handle

    def get_mtime(self, src):
        path = pathlib.Path(src).resolve()
        with ExpectedExceptionContext(OSError):
            mtime = path.stat().st_mtime
        return datetime.datetime.fromtimestamp(mtime, datetime.timezone.utc)



#!/usr/bin/env python3
"""
Test regex module.
"""

import datetime
import unittest

from dalood.regex import (
    PatternType,
    RegexPattern,
    GlobPattern,
    LiteralPattern,
    get_regex,
)


class TestGetRegex(unittest.TestCase):
    """
    Test get_regex function.
    """

    def test_passthrough(self):
        """
        RegexPattern instances are returned.
        """
        for pattern in (
            RegexPattern(r".*\.txt$"),
            GlobPattern(r"*.json"),
            LiteralPattern("foo.yaml"),
        ):
            self.assertIs(pattern, get_regex(pattern))

    def test_regex_type(self):
        """
        RegexPatterns are returned.
        """
        pattern = get_regex("f.o", pattern_type="regex")
        self.assertIs(pattern.__class__, RegexPattern)
        self.assertEqual(pattern.regex.pattern, r"f.o")

    def test_glob_type(self):
        """
        GlobPatterns are returned.
        """
        pattern = get_regex("f*o", pattern_type="glob")
        self.assertIs(pattern.__class__, GlobPattern)
        self.assertEqual(pattern.regex.pattern, r"(?s:f.*o)\Z")

    def test_literal_type(self):
        """
        LiteralPatterns are returned.
        """
        pattern = get_regex("f?o", pattern_type="literal")
        self.assertIs(pattern.__class__, LiteralPattern)
        self.assertEqual(pattern.regex.pattern, r"f\?o")

    def test_invalid_pattern_types(self):
        """
        Invalid pattern types raise errors.
        """
        for invalid_type in ("globx", None, 5, False):
            with self.subTest(invalid_type=invalid_type), self.assertRaises(ValueError):
                get_regex("foo", pattern_type=invalid_type)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Test JSON loaders.
"""

import json
import unittest

from dalood.exception import LoaderValueError
from dalood.loader.json import JSONFileLoader, JSONUrlLoader
from dalood.manager import Manager


from test_dalood.utils import tmp_file


JSON_EXT = ".json"


def _get_json():
    """
    Get data and its JSON representation.
    """
    data = {"foo": 1, "bar": 2, "items": [3, 4, 5], "bool": True}
    return data, json.dumps(data)


class TestJSON(unittest.TestCase):
    """
    Test JSON loading.
    """

    def setUp(self):
        self.data, self.json_txt = _get_json()
        self.tmp_file_ctx = tmp_file(text=self.json_txt, extension=JSON_EXT)

    def test_json_file_load(self):
        """
        JSON is loaded from files.
        """
        man = Manager()
        man.register_loader(rf"^.*\{JSON_EXT}$", JSONFileLoader())
        with self.tmp_file_ctx as (_, _, path):
            self.assertEqual(self.data, man.get(path))

    def test_uri_file_load(self):
        """
        JSON is loaded from URIs.
        """
        man = Manager()
        man.register_loader(r"^file://.*$", JSONUrlLoader())
        with self.tmp_file_ctx as (_, _, path):
            self.assertEqual(self.data, man.get(path.as_uri()))

    def test_patterns(self):
        """
        JSON file loader registers its own patterns.
        """
        man = Manager()
        loader = JSONFileLoader()
        with self.tmp_file_ctx as (_, _, path):
            with self.assertRaises(LoaderValueError):
                man.get(path.as_uri())

            loader.register_patterns(man)
            self.assertEqual(self.data, man.get(path))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Test Manager.
"""

import datetime
import unittest
import re
from unittest.mock import MagicMock

from dalood.loader.text import TextFileLoader
from dalood.manager import Manager
from dalood.regex import PatternType, GlobPattern


from test_dalood.utils import tmp_file


TEXT_EXT = ".txt"


def shift_times(manager):
    """
    Shift load times one day into the past and access times 6 hours into the
    past for all loaded data in the manager.

    Args:
        manager:
            A Manager instance.
    """
    load_delta = datetime.timedelta(days=1)
    access_delta = datetime.timedelta(hours=6)
    for wrapper in manager._cache.values():  # pylint: disable=protected-access
        wrapper.load_time -= load_delta
        wrapper.access_time -= access_delta


class TestManager(unittest.TestCase):
    """
    Test Manager class.
    """

    def setUp(self):
        self.text = "placeholder text"
        self.tmp_file_ctx = tmp_file(text=self.text, extension=TEXT_EXT)

        self.man = Manager()
        self.loader = TextFileLoader()
        self.man.register_loader(rf"^.*\{TEXT_EXT}$", self.loader)

    def _pattern_iterator(self):
        """
        Common convenience method for testing pattern types.
        """
        with self.tmp_file_ctx as (text, mtime, path):
            for name, pattern in (
                ("glob", f"*{TEXT_EXT}"),
                ("literal", str(path)),
                ("regex", rf"^.*{re.escape(TEXT_EXT)}$"),
            ):
                yield text, mtime, path, name, pattern

    def test_register_loader_with_x_pattern(self):
        """
        Convenience wrappers around register_loader work.
        """
        for text, _, path, name, pattern in self._pattern_iterator():
            with self.subTest(name=name, pattern=pattern):
                self.man.loaders.clear()
                getattr(self.man, f"register_loader_with_{name}_pattern")(
                    pattern, TextFileLoader()
                )
                self.assertEqual(text, self.man.get(path))

    def test_get_loader(self):
        """
        Manager returns loader.
        """
        self.assertEqual(self.loader, self.man.get_loader(f"foo{TEXT_EXT}"))

    def test_clear_cache(self):
        """
        Manager clears cached data.
        """
        with self.tmp_file_ctx as (_, _, path):
            # Clear all.
            self.man.get(path)
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache()
            self.assertEqual([], list(self.man))

            # Clear by load time.
            self.man.get(path)
            shift_times(self.man)
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache(age={"hours": 36})
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache(age={"hours": 20})
            self.assertEqual([], list(self.man))

            # Clear by access time.
            self.man.get(path)
            shift_times(self.man)
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache(age={"hours": 20}, by_access_time=True)
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache(age={"hours": 3}, by_access_time=True)
            self.assertEqual([], list(self.man))

            # Clear by pattern.
            self.man.get(path)
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache(pattern="sfdsdfsdf", pattern_type=PatternType.LITERAL)
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache(pattern=f"*{TEXT_EXT}", pattern_type=PatternType.GLOB)
            self.assertEqual([], list(self.man))

            self.man.get(path)
            self.assertEqual([str(path)], list(self.man))
            self.man.clear_cache(pattern=GlobPattern(f"*{TEXT_EXT}"))
            self.assertEqual([], list(self.man))

    def test_clear_cache_with_x_pattern(self):
        """
        Convenience wrappers around clear_cache work.
        """
        for _, _, path, name, pattern in self._pattern_iterator():
            with self.subTest(name=name, pattern=pattern):
                self.man.get(path)
                self.assertEqual([str(path)], list(self.man))
                getattr(self.man, f"clear_cache_with_{name}_pattern")(pattern)
                self.assertEqual([], list(self.man))

    def test_refresh(self):
        """
        Manager refreshes data.
        """
        with self.tmp_file_ctx as (_, _, path):
            self.man.get(path)
            shift_times(self.man)
            wrapper = self.man._get_wrapper(path)  # pylint: disable=protected-access
            prev_load_time = wrapper.load_time
            prev_access_time = wrapper.access_time
            self.man.refresh()
            wrapper = self.man._get_wrapper(path)  # pylint: disable=protected-access
            self.assertGreater(wrapper.load_time, prev_load_time)
            self.assertGreater(wrapper.access_time, prev_access_time)

            # Refresh by pattern.
            shift_times(self.man)
            wrapper = self.man._get_wrapper(path)  # pylint: disable=protected-access
            prev_load_time = wrapper.load_time
            prev_access_time = wrapper.access_time
            self.man.refresh(pattern="sfdsdfsdf", pattern_type=PatternType.LITERAL)
            wrapper = self.man._get_wrapper(path)  # pylint: disable=protected-access
            self.assertEqual(wrapper.load_time, prev_load_time)
            self.assertEqual(wrapper.access_time, prev_access_time)

            self.man.refresh(pattern=f"*{TEXT_EXT}", pattern_type=PatternType.GLOB)
            wrapper = self.man._get_wrapper(path)  # pylint: disable=protected-access
            self.assertGreater(wrapper.load_time, prev_load_time)
            self.assertGreater(wrapper.access_time, prev_access_time)

    def test_refresh_with_x_pattern(self):
        """
        Convenience wrappers around refresh work.
        """
        for _, _, path, name, pattern in self._pattern_iterator():
            with self.subTest(name=name, pattern=pattern):
                self.man.get(path)
                shift_times(self.man)
                wrapper = self.man._get_wrapper(
                    path
                )  # pylint: disable=protected-access
                prev_load_time = wrapper.load_time
                prev_access_time = wrapper.access_time
                getattr(self.man, f"refresh_with_{name}_pattern")(pattern)
                wrapper = self.man._get_wrapper(
                    path
                )  # pylint: disable=protected-access
                self.assertGreater(wrapper.load_time, prev_load_time)
                self.assertGreater(wrapper.access_time, prev_access_time)

    def test_get(self):
        """
        Manager returns data.
        """
        with self.tmp_file_ctx as (text, _, path):
            self.assertEqual(text, self.man.get(path))

            wrapper = self.man._get_wrapper(path)  # pylint: disable=protected-access

            # Reload parameter.
            wrapper.load = MagicMock(side_effect=wrapper.load)
            self.man.get(path)
            wrapper.load.assert_not_called()
            self.man.get(path, reload=True)
            wrapper.load.assert_called()

            # Refresh parameter.
            wrapper.refresh = MagicMock(side_effect=wrapper.refresh)
            self.man.get(path)
            wrapper.refresh.assert_not_called()
            self.man.get(path, refresh=True)
            wrapper.refresh.assert_called()

    def test_get_mtime(self):
        """
        Manager returns last modification time of source.
        """
        with self.tmp_file_ctx as (_, mtime, path):
            self.assertEqual(mtime, self.man.get_mtime(path))


if __name__ == "__main__":
    unittest.main()
